//
// Rep.java - Repository
//
// The Repository
//   contains a set of files and tombstones
//   manages files and tombstones below a specified base directory
//   has a location value, used when adopting files
//
// Each file or tombstone
//   cannot be named starting with .
//   is identified by a hierarchical name, eg: dir1/dir2/filename
//   has a location value and timestamp
//   has a version too (not visible to clients)
//
// Directories
//   are not explicitly created
//   exist for as long as there are files in them
//

package nyangau.rep;

import java.util.*;
import java.util.logging.*;
import java.io.*;

public class Rep
  {
  private static Logger logger = Logger.getLogger(Rep.class.getName());
//...sreadMeta:2:
protected static Properties readMeta(File f)
  throws FileNotFoundException, IOException
  {
  Properties p = new Properties();
  FileInputStream fis = new FileInputStream(f);
  p.load(fis);
  fis.close();
  return p;
  }
//...e
//...swriteMeta:2:
protected static void writeMeta(File f, Properties p)
  throws FileNotFoundException, IOException
  {
  FileOutputStream fos = new FileOutputStream(f);
  p.store(fos, null);
  fos.close();
  }
//...e
//...sconstants:2:
public static final String SEP = "/";
protected static final String FNPREF_META      = ".m_";
protected static final String FNPREF_METAFIRST = ".f_";
protected static final String FNPREF_METANEW   = ".n_";
protected static final String FNPREF_DATA      = ".";
protected static final String META_LOC = "loc";
protected static final String META_VER = "ver";
protected static final String META_TS  = "ts";
//...e
  protected Object oSync;
  protected File base;
  protected int loc;
//...sRep:2:
public Rep(String basedir, int loc)
  throws FileNotFoundException, IOException
  {
  oSync = new Object();
  base = new File(basedir);
  if ( ! base.isDirectory() )
    throw new FileNotFoundException(basedir+" is not a directory");
  this.loc = loc;
  recover(base);
  }
//...e
//...snewMeta:2:
protected static void newMeta(
  int loc, long ver, long ts,
  File fMetaNew, File fMeta
  )
  throws FileNotFoundException, IOException
  {
  Properties p = new Properties();
  p.setProperty(META_LOC, Integer.toString(loc));
  if ( ver != -1 )
    p.setProperty(META_VER, Long.toString(ver));
  p.setProperty(META_TS, Long.toString(ts));
  writeMeta(fMetaNew, p);
  fMeta.delete();
  fMetaNew.renameTo(fMeta);
  }
//...e
//...srecover:2:
protected void recover(File dir)
  throws FileNotFoundException, IOException
  {
  String[] fns;
  Set<String> fnsData = new HashSet<String>();
  Set<String> fnsOther = new HashSet<String>();
  // Roll back or forward metadata creates or updates
  // Remember data files
  // Remember other files
  fns = dir.list();
  for ( int i = 0; i < fns.length; i++ )
    {
    String fn = fns[i];
    File f = new File(dir, fn);
    if ( fn.startsWith(FNPREF_META) )
      ;
    else if ( fn.startsWith(FNPREF_METAFIRST) )
      {
      logger.info("Rep: roll back metadata creation, delete "+f.getName());
      f.delete();
      }
    else if ( fn.startsWith(FNPREF_METANEW) )
      {
      File fMeta = new File(dir, FNPREF_META+fn.substring(FNPREF_METANEW.length()));
      if ( fMeta.exists() )
        {
        logger.info("Rep: roll back metadata update, keep "+fMeta.getName()+" and delete "+f.getName());
        f.delete();
        }
      else
        {
        logger.info("Rep: roll forward metadata update, rename "+f.getName()+" to "+fMeta.getName());
        f.renameTo(fMeta);
        }
      }
    else if ( fn.startsWith(FNPREF_DATA) )
      fnsData.add(fn);
    else if ( f.isFile() )
      fnsOther.add(fn);
    }
  // Forget any data files referenced by metadata
  fns = dir.list();
  for ( int i = 0; i < fns.length; i++ )
    {
    String fn = fns[i];
    if ( fn.startsWith(FNPREF_META) )
      {
      File fMeta = new File(dir, fn);
      Properties p = readMeta(fMeta);
      String ver = p.getProperty(META_VER);
      if ( ver != null )
        {
        String fnData = FNPREF_DATA+ver+"_"+fn.substring(FNPREF_META.length());
        if ( ! fnsData.remove(fnData) )
          {
          File fData = new File(dir, fnData);
          logger.warning("Rep: corrupted filesystem, "+fData.getName()+" is missing");
          }
        }
      }
    }
  // Delete data files not referenced by metadata
  for ( String fn : fnsData )
    {
    File fOrphan = new File(dir, fn);
    logger.info("Rep: deleting orphan "+fOrphan.getName());
    fOrphan.delete();
    }
  // Adopt any other files we find
  for ( String fn : fnsOther )
    {
    File fOther     = new File(dir,                  fn);
    File f          = new File(dir, FNPREF_DATA+"0_"+fn);
    File fMetaFirst = new File(dir, FNPREF_METAFIRST+fn);
    File fMeta      = new File(dir, FNPREF_META     +fn);
    logger.info("Rep: adopting "+fOther.getName()+" as "+f.getName());
    fOther.renameTo(f);
    newMeta(loc, 0, 0, fMetaFirst, fMeta);
    }
  // Do any nested directories
  fns = dir.list();
  for ( int i = 0; i < fns.length; i++ )
    {
    String fn = fns[i];
    File f = new File(dir, fn);
    if ( f.isDirectory() )
      recover(f);
    }
  }
//...e
//...ssplit:2:
protected static Vector<String> split(String name)
  {
  StringTokenizer st = new StringTokenizer(name, SEP);
  Vector<String> v = new Vector<String>();
  while ( st.hasMoreTokens() )
    v.add(st.nextToken());
  return v;
  }
//...e
//...sprefix:2:
protected static String prefix(Vector<String> v, int nV)
  {
  StringBuffer sb = new StringBuffer();
  for ( int i = 0; i < nV; i++ )
    {
    sb.append(v.elementAt(i));
    sb.append(File.separator);
    }
  return sb.toString();
  }
//...e
//...shasContent:2:
// Unfortunately this can be quite expensive.
// Only use when necessary.

public boolean hasContent(File dir)
  throws FileNotFoundException, IOException
  {
  String fns[] = dir.list();
  for ( int i = 0; i < fns.length; i++ )
    {
    String fn = fns[i];
    if ( fn.startsWith(FNPREF_META) )
      {
      File fMeta = new File(dir, fn);
      Properties p = readMeta(fMeta);
      if ( p.getProperty(META_VER) != null )
        return true;
      }
    }
  for ( int i = 0; i < fns.length; i++ )
    {
    File ndir = new File(dir, fns[i]);
    if ( ndir.isDirectory() )
      if ( hasContent(ndir) )
        return true;
    }
  return false;
  }
//...e
//...scalcSearchDir:2:
protected File calcSearchDir(String name)
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    return base;
  else
    {    
    String dn = prefix(v, v.size()-1);
    String bn = v.elementAt(v.size()-1);
    return new File(base, dn+bn);
    }
  }
//...e
//...sisFile:2:
// Return true if named file exists

public boolean isFile(String name)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    return false;
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      return false;
    Properties p = readMeta(fMeta);
    String ver = p.getProperty(META_VER);
    return ver != null;
/*
    File fData = new File(base, dn+FNPREF_DATA+ver+"_"+bn);
    return fData.exists();
*/
    }
  }
//...e
//...sisTombstone:2:
// Return true if named tombstone exists

public boolean isTombstone(String name)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    return false;
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      return false;
    Properties p = readMeta(fMeta);
    String ver = p.getProperty(META_VER);
    return ver == null;
/*
    File fData = new File(base, dn+FNPREF_DATA+ver+"_"+bn);
    return ! fData.exists();
*/
    }
  }
//...e
//...sisDirectory:2:
public boolean isDirectory(String name)
  throws FileNotFoundException, IOException
  {
  synchronized ( oSync )
    {
    File dir = calcSearchDir(name);
    return dir.isDirectory() && hasContent(dir);
    }
  }
//...e
//...sisDirectoryEvenEmpty:2:
public boolean isDirectoryEvenEmpty(String name)
  throws FileNotFoundException, IOException
  {
  synchronized ( oSync )
    {
    File dir = calcSearchDir(name);
    return dir.isDirectory();
    }
  }
//...e
//...sgetLoc:2:
public int getLoc(String name)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    throw new FileNotFoundException(name+" is not a valid file name");
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      throw new FileNotFoundException(name+" not in repository");
    Properties p = readMeta(fMeta);
    String locS = p.getProperty(META_LOC);
    return Integer.parseInt(locS);
    }
  }
//...e
//...sgetTs:2:
public long getTs(String name)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    throw new FileNotFoundException(name+" is not a valid file name");
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      throw new FileNotFoundException(name+" not in repository");
    Properties p = readMeta(fMeta);
    String tsS = p.getProperty(META_TS);
    return Long.parseLong(tsS);
    }
  }
//...e
//...sclass RepMetaData:2:
public class RepMetaData
  {
  public int loc;
  public long ts;
  public boolean isFile;
  public RepMetaData(int loc, long ts, boolean isFile)
    { this.loc = loc; this.ts = ts; this.isFile = isFile; }
  }
//...e
//...sgetMetaData:2:
// This is a special method for use by SharedRep when starting up

public RepMetaData getMetaData(String name)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    return null;
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      return null;
    Properties p = readMeta(fMeta);
    String locS = p.getProperty(META_LOC);
    String tsS = p.getProperty(META_TS);
    return new RepMetaData(
      Integer.parseInt(locS),
      Long.parseLong(tsS),
      p.getProperty(META_VER) != null
      );
    }
  }
//...e
//...sgetLength:2:
protected long getLength(String name)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    throw new FileNotFoundException(name+" is not a valid file name");
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      throw new FileNotFoundException(name+" not in repository");
    Properties p = readMeta(fMeta);
    String verS = p.getProperty(META_VER);
    if ( verS == null )
      throw new FileNotFoundException(name+" is a tombstone");
    File fData = new File(base, dn+FNPREF_DATA+verS+"_"+bn);
    return fData.length();
    }
  }
//...e
//...sfilesInDirectory:2:
public String[] filesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  File dir = calcSearchDir(name);
  if ( ! dir.isDirectory() )
    return new String[0];
  String fns[] = dir.list();
  Vector<String> rf = new Vector<String>();
  synchronized ( oSync )
    {
    for ( int i = 0; i < fns.length; i++ )
      {
      String fn = fns[i];
      if ( fn.startsWith(FNPREF_META) )
        {
        File fMeta = new File(dir, fn);
        Properties p = readMeta(fMeta);
        if ( p.getProperty(META_VER) != null )
          rf.add(fn.substring(FNPREF_META.length()));
        }
      }
    }
  return rf.toArray(new String[0]);
  }
//...e
//...stombstonesInDirectory:2:
protected String[] tombstonesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  File dir = calcSearchDir(name);
  if ( ! dir.isDirectory() )
    return new String[0];
  String fns[] = dir.list();
  Vector<String> rf = new Vector<String>();
  synchronized ( oSync )
    {
    for ( int i = 0; i < fns.length; i++ )
      {
      String fn = fns[i];
      if ( fn.startsWith(FNPREF_META) )
        {
        File fMeta = new File(dir, fn);
        Properties p = readMeta(fMeta);
        if ( p.getProperty(META_VER) == null )
          rf.add(fn.substring(FNPREF_META.length()));
        }
      }
    }
  return rf.toArray(new String[0]);
  }
//...e
//...sdirectoriesInDirectory:2:
public String[] directoriesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  File dir = calcSearchDir(name);
  if ( ! dir.isDirectory() )
    return new String[0];
  String fns[] = dir.list();
  Vector<String> rf = new Vector<String>();
  synchronized ( oSync )
    {
    for ( int i = 0; i < fns.length; i++ )
      {
      String fn = fns[i];
      File f = new File(dir, fn);
      if ( f.isDirectory() && hasContent(f) )
        rf.add(fn);
      }
    }
  return rf.toArray(new String[0]);
  }
//...e
//...sreadFile:2:
// Read file
// We can ask for specific loc and/or ts

public byte[] readFile(String name, int locWanted, long tsWanted)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    throw new FileNotFoundException(name+" is not a valid file name");
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( ! fMeta.exists() )
      throw new FileNotFoundException(name+" not in repository");
    Properties p = readMeta(fMeta);
    String verS = p.getProperty(META_VER);
    if ( verS == null )
      throw new FileNotFoundException(name+" is a tombstone");
    if ( locWanted != -1 &&
         locWanted != Integer.parseInt(p.getProperty(META_LOC)) )
      throw new FileNotFoundException(name+" has wrong loc");
    if ( tsWanted != -1 &&
         tsWanted != Long.parseLong(p.getProperty(META_TS)) )
      throw new FileNotFoundException(name+" has wrong ts");
    File fData = new File(base, dn+FNPREF_DATA+verS+"_"+bn);
    long l = fData.length();
    byte[] b = new byte[(int)l];
    FileInputStream is = new FileInputStream(fData);
    is.read(b);
    is.close();
    return b;
    }
  }

public byte[] readFile(String name)
  throws FileNotFoundException, IOException
  {
  return readFile(name, -1, -1);
  }
//...e
//...swriteFile:2:
// Atomically set file to new content, annotated with loc and ts.
// Optionally overwrite previous file content.

//...simportData:2:
protected void importData(byte[] b, File f)
  throws FileNotFoundException, IOException
  {
  FileOutputStream fos = new FileOutputStream(f);
  fos.write(b);
  fos.close();
  }
//...e

public void writeFile(
  String name, int loc, long ts, boolean overwrite, byte[] b
  )
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    throw new FileNotFoundException(name+" is not a valid file name");
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( fMeta.exists() )
      {
      Properties p = readMeta(fMeta);
      String verS = p.getProperty(META_VER);
      if ( verS != null )
        {
        if ( !overwrite )
          throw new FileNotFoundException(name+" cannot be overwritten");
        // overwrite a file
        long ver = Long.parseLong(verS);
        File fData    = new File(base, dn+FNPREF_DATA+Long.toString(ver  )+"_"+bn);
        File fDataNew = new File(base, dn+FNPREF_DATA+Long.toString(ver+1)+"_"+bn);
        File fMetaNew = new File(base, dn+FNPREF_METANEW                      +bn);
        importData(b, fDataNew);
        newMeta(loc, ver+1, ts, fMetaNew, fMeta);
        fData.delete();
        }
      else
        // overwrite a tombstone
        {
        File fDataNew = new File(base, dn+FNPREF_DATA+"0_"+bn);
        File fMetaNew = new File(base, dn+FNPREF_METANEW  +bn);
        importData(b, fDataNew);
        newMeta(loc, 0, ts, fMetaNew, fMeta);
        }
      }
    else
      // write a new file
      {
      if ( v.size() != 1 )
        {
        String dnParent = prefix(v, v.size()-2);
        String bnParent = v.elementAt(v.size()-2);
        File dirParent = new File(base, dnParent+bnParent);
        dirParent.mkdirs(); // fails if already exist, so ignore return
        }
      File fDataNew   = new File(base, dn+FNPREF_DATA+"0_"+bn);
      File fMetaFirst = new File(base, dn+FNPREF_METAFIRST+bn);
      importData(b, fDataNew);
      newMeta(loc, 0, ts, fMetaFirst, fMeta);
      }
    }
  }
//...e
//...sdeleteFile:2:
// Atomically set file to a tombstone, annotated with loc and ts

protected void deleteFile(String name, int loc, long ts)
  throws FileNotFoundException, IOException
  {
  Vector<String> v = split(name);
  if ( v.size() == 0 )
    throw new FileNotFoundException(name+" is not a valid file name");
  String dn = prefix(v, v.size()-1);
  String bn = v.elementAt(v.size()-1);
  File fMeta = new File(base, dn+FNPREF_META+bn);
  synchronized ( oSync )
    {
    if ( fMeta.exists() )
      {
      Properties p = readMeta(fMeta);
      String verS = p.getProperty(META_VER);
      if ( verS != null )
        {
        // overwrite a file with a tombstone
        long ver = Long.parseLong(verS);
        File fData    = new File(base, dn+FNPREF_DATA+Long.toString(ver)+"_"+bn);
        File fMetaNew = new File(base, dn+FNPREF_METANEW                    +bn);
        newMeta(loc, -1, ts, fMetaNew, fMeta);
        fData.delete();
        }
      else
        // overwrite a tombstone with a new tombstone
        {
        File fMetaNew = new File(base, dn+FNPREF_METANEW+bn);
        newMeta(loc, -1, ts, fMetaNew, fMeta);
        }
      }
    else
      // write new tombstone
      {
      if ( v.size() != 1 )
        {
        String dnParent = prefix(v, v.size()-2);
        String bnParent = v.elementAt(v.size()-2);
        File dirParent = new File(base, dnParent+bnParent);
        dirParent.mkdirs(); // fails if already exist, so ignore return
        }
      File fMetaFirst = new File(base, dn+FNPREF_METAFIRST+bn);
      newMeta(loc, -1, ts, fMetaFirst, fMeta);
      }
    }
  }
//...e
//...sdump:2:
protected void dump(String name, String indent)
  throws FileNotFoundException, IOException
  {
  if ( isFile(name) )
    System.out.println(indent+"file "+name+" loc="+getLoc(name)+" ts="+getTs(name)+" length="+getLength(name));
  else if ( isTombstone(name) )
    System.out.println(indent+"tombstone "+name+" loc="+getLoc(name)+" ts="+getTs(name));
  else if ( isDirectoryEvenEmpty(name) )
    {
    System.out.println(indent+"directory "+name);
    String[] fns = filesInDirectory(name);
    for ( int i = 0; i < fns.length; i++ )
      dump(name+SEP+fns[i], indent+"  ");
    String[] tns = tombstonesInDirectory(name);
    for ( int i = 0; i < tns.length; i++ )
      dump(name+SEP+tns[i], indent+"  ");
    String[] dns = directoriesInDirectory(name);
    for ( int i = 0; i < dns.length; i++ )
      dump(name+SEP+dns[i], indent+"  ");
    }
  else
    System.out.println(indent+"? "+name);
  }

public void dump()
  throws FileNotFoundException, IOException
  {
  dump("", "");
  }
//...e
//...sdebugCommand:2:
public boolean debugCommand(String[] args)
  throws FileNotFoundException, IOException
  {
  if ( args[0].equals("isFile") )
    System.out.println(isFile(args[1]));
  else if ( args[0].equals("isTombstone") )
    System.out.println(isTombstone(args[1]));
  else if ( args[0].equals("getLoc") )
    System.out.println(getLoc(args[1]));
  else if ( args[0].equals("getTs") )
    System.out.println(getTs(args[1]));
  else if ( args[0].equals("getLength") )
    System.out.println(getLength(args[1]));
  else if ( args[0].equals("readFile") )
    {
    byte[] b = readFile(args[1]);
    System.out.write(b);
    }
  else if ( args[0].equals("writeFile") )
    {
    byte[] b = (args[3]+"\n").getBytes("UTF-8");
    writeFile(args[1], Integer.parseInt(args[2]), System.currentTimeMillis(), true, b);
    }
  else if ( args[0].equals("deleteFile") )
    deleteFile(args[1], Integer.parseInt(args[2]), System.currentTimeMillis());
  else if ( args[0].equals("dump") )
    dump();
  else if ( args[0].equals("help") )
    {
    System.out.println("isFile name");
    System.out.println("isTombstone name");
    System.out.println("getLoc name");
    System.out.println("getTs name");
    System.out.println("readFile name");
    System.out.println("writeFile name loc data");
    System.out.println("deleteFile name loc");
    System.out.println("dump");
    }
  else
    return false;
  return true;
  }
//...e
//...sdebugDriver:2:
// Interactive test driver

public void debugDriver()
  throws FileNotFoundException, IOException
  {
  InputStreamReader isr = new InputStreamReader(System.in);
  BufferedReader br = new BufferedReader(isr);
  String line;
  while ( (line = br.readLine()) != null )
    {
    StringTokenizer st = new StringTokenizer(line);
    Vector<String> v = new Vector<String>();
    while ( st.hasMoreTokens() )
      v.add(st.nextToken());
    if ( v.size() > 0 )
      {
      String[] a = v.toArray(new String[0]);
      try
        {
        if ( ! debugCommand(a) )
          System.out.println("Not understood, use help");
        }
      catch ( Exception e )
        {
        e.printStackTrace();
        }
      }
    }
  }
//...e
//...smain:2:
public static void main(String[] args)
  {
  try
    {
    Rep rep = new Rep(
      args[0],                    // basedir
      Integer.parseInt(args[1])   // loc
      );
    rep.debugDriver();
    }
  catch ( Exception e )
    {
    e.printStackTrace();
    }
  }
//...e
  }
