//
// SharedRepClient.java - Simple client of Shared Repository
//
// Protocol spoken
//
//   pw -> rc
//
//   isFile:                 'f' name           -> rc
//   isDirectory:            'd' name           -> rc
//   filesInDirectory:       'F' name           -> rc #names name{#names}
//   directoriesInDirectory: 'D' name           -> rc #names name{#names}
//   readFile                'R' name           -> rc #b b{#b}
//   writeFile:              'W' name #b b{#b}  -> rc
//   deleteFile:             'Z' name           -> rc
//   findXPath:              'X' name f r c x n -> rc #b b{#b}
//

package nyangau.rep;

//...simports:0:
import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import javax.net.*;
import javax.net.ssl.*;
import javax.security.auth.x500.*;
import javax.xml.xpath.XPathExpressionException;
//...e

public class SharedRepClient
  {
  protected String host;
  protected int    port;
  protected String password;
  protected SocketFactory sf;
  protected Object oSyncPool;
  protected LinkedList<Connection> pool;
  protected Reaper reaper;
//...sSharedRepClient:2:
//...sgetProp:2:
protected static String getProp(Map<String,Object> env, String p, String def)
  {
  String v;
  if ( env != null && (v = (String) env.get(p)) != null )
    return v;
  if ( (v = System.getProperty(p)) != null )
    return v;
  return def;
  }
//...e

public SharedRepClient(
  String connection,
  String password,
  Map<String,Object> env
  )
  throws FileNotFoundException, IOException
  {
  if ( ! connection.startsWith("srep://") )
    throw new IllegalArgumentException("connection should be of the form srep://host:port");
  connection = connection.substring(7);
  int i = connection.lastIndexOf(":");
  if ( i != -1 )
    {
    host = connection.substring(0, i);
    port = Integer.parseInt(connection.substring(i+1));
    }
  else
    {
    host = connection;
    port = 6600;
    }
  this.password = password;

  // Setup SSL stuff
  String ksfn = getProp(env, "nyangau.rep.truststore", "SharedRepClient.jks");
  String kspw = getProp(env, "nyangau.rep.storepass" , "password"           );
  try
    {
    KeyStore ks = KeyStore.getInstance("JKS");
    char[] ksp = kspw.toCharArray();
    ks.load(new FileInputStream(ksfn), ksp);
    KeyManagerFactory kmf = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
    kmf.init(ks, ksp);
    KeyStore ts = KeyStore.getInstance("JKS");
    char[] tsp = kspw.toCharArray();
    ts.load(new FileInputStream(ksfn), tsp);
    TrustManagerFactory tmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
    tmf.init(ts);
    SSLContext sc = SSLContext.getInstance("TLS");
    sc.init(kmf.getKeyManagers(), tmf.getTrustManagers(), null);
    sf = sc.getSocketFactory();
    }
  catch ( GeneralSecurityException e )
    {
    throw new IOException("Couldn't set up the SSL socket factory");
    }

  oSyncPool = new Object();
  pool = new LinkedList<Connection>();
  reaper = null;
  }

public SharedRepClient(String connection, String password)
  throws FileNotFoundException, IOException
  {
  this(connection, password, null);
  }
//...e
//...sclass Connection:2:
//...sgetCN:2:
protected static String getCN(String name)
  throws IOException
  {
  StringTokenizer st = new StringTokenizer(name, ",");
  while ( st.hasMoreTokens() )
    {
    String s = st.nextToken();
    int i = s.indexOf("CN=");
    if ( i >= 0 )
      return s.substring(i+3);
    }
  throw new IOException("DN has no CN");
  }
//...e
//...scheckCertificate:2:
protected static void checkCertificate(String dns, java.security.cert.X509Certificate cert)
  throws IOException
  {
  String cn = getCN( cert.getSubjectDN().getName() );
  if ( dns.equals(cn) )
    return;
  try
    {
    Collection c = cert.getSubjectAlternativeNames();
    if ( c != null )
      {
      Iterator i = c.iterator();
      while ( i.hasNext() )
        {
        List l = (List) i.next();
        if ( ((Integer)l.get(0)).intValue() == 2 &&
             dns.equals((String)l.get(1)) )
          return;
        }
      }
    throw new IOException("DNS does not match CN in certificate DN or alternate subject DNS names");
    }
  catch ( CertificateParsingException e )
    {
    throw new IOException("Error parsing certificate");
    }
  }
//...e
//...scheckSession:2:
protected static void checkSession(SSLSession ssess)
  throws IOException
  {
  try
    {
    java.security.cert.Certificate[] certs = ssess.getPeerCertificates();
    if ( certs == null || certs.length == 0 )
      throw new IOException("No certificates");
    if ( ! ( certs[0] instanceof java.security.cert.X509Certificate ) )
      throw new IOException("Expected X509 certificate");
    String dns = ssess.getPeerHost();
    checkCertificate(dns, (java.security.cert.X509Certificate) certs[0]);
    }
  catch ( SSLPeerUnverifiedException e )
    {
    throw new IOException("Cannot establish trust in certificates");
    }
  }
//...e

public class Connection
  {
  protected Socket s;
  public DataOutputStream dos;
  public DataInputStream  dis;
  public long lastUsed;
  public Connection()
    throws IOException
    {
    s = sf.createSocket(host, port);

    String p = System.getProperty("nyangau.rep.verifyHostname");
    if ( p != null && p.equalsIgnoreCase("false") )
      // The user has explicitly ask that we don't check that
      // the certificate DN (or subjectAltNames) matched the DNS.
      // These must be checked to ensure a secure connection.
      ;
    else
      // We'd better check properly.
      {
      SSLSocket ss = (SSLSocket) s;
      SSLSession ssess = ss.getSession();
      checkSession(ssess);
      }

    dis = new DataInputStream (s.getInputStream ());
    dos = new DataOutputStream(s.getOutputStream());
    }
  public void disconnect()
    throws IOException
    {
    dis.close();
    dos.close();
    s.close();
    }
  }
//...e
//...sclass Reaper:2:
public class Reaper extends Thread
  {
  public static final long T_LINGER = 10000;
  public static final long T_CHECK  =  5000;
  public void run()
    {
    for ( ;; )
      {
      try
        {
        sleep(T_CHECK);
        }
      catch ( InterruptedException e )
        {
        }
      synchronized ( oSyncPool )
        {
        long t = System.currentTimeMillis();
        LinkedList<Connection> poolNew = new LinkedList<Connection>();
        while ( ! pool.isEmpty() )
          {
          Connection c = pool.removeFirst();
          if ( c.lastUsed + T_LINGER < t )
            try
              {
              c.disconnect();
              }
            catch ( IOException e )
              {
              }
          else
            poolNew.addLast(c);
          }
        pool = poolNew;
        if ( pool.isEmpty() )
          {
          reaper = null;
          return; // Reaper ends
          }
        }
      }
    }
  }
//...e
//...sacquireConnection:2:
// Throws IOException if cannot connect
// Throws IOException if authentication error
// Returns connection otherwise

protected Connection acquireConnection()
  throws IOException
  {
  synchronized ( oSyncPool )
    {
    if ( ! pool.isEmpty() )
      return pool.removeFirst();
    }
  Connection c = null;
  try
    {
    c = new Connection();
    c.dos.writeUTF(password);
    c.dos.flush();
    if ( c.dis.readUnsignedByte() == 'S' )
      {
      c.disconnect();
      throw new IOException("authentication failed");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      try
        {
        c.disconnect();
        }
      catch ( IOException e2 )
        {
        }
    throw e;
    }
  return c;
  }
//...e
//...sreleaseConnection:2:
protected void releaseConnection(Connection c)
  {
  c.lastUsed = System.currentTimeMillis();
  synchronized ( oSyncPool )
    {
    pool.addLast(c);
    if ( reaper == null )
      {
      reaper = new Reaper();
      reaper.setDaemon(true);
      reaper.start();
      }
    }
  }
//...e

//...sisFile:2:
public boolean isFile(String name)
  throws IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('f');
    c.dos.writeUTF(name);
    c.dos.flush();
    int rc = c.dis.readUnsignedByte();
    releaseConnection(c);
    return rc == 'f';
    }
  catch ( IOException e )
    {
    c.disconnect();
    throw e;
    }
  }
//...e
//...sisDirectory:2:
public boolean isDirectory(String name)
  throws IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('d');
    c.dos.writeUTF(name);
    c.dos.flush();
    int rc = c.dis.readUnsignedByte();
    releaseConnection(c);
    return rc == 'd';
    }
  catch ( IOException e )
    {
    c.disconnect();
    throw e;
    }
  }
//...e
//...sfilesInDirectory:2:
public String[] filesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('F');
    c.dos.writeUTF(name);
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'F':
        {
        Vector<String> v = new Vector<String>();
        int n = c.dis.readInt();
        for ( int i = 0; i < n; i++ )
          v.add(c.dis.readUTF());
        releaseConnection(c);
        return v.toArray(new String[0]);
        }
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...sdirectoriesInDirectory:2:
public String[] directoriesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('D');
    c.dos.writeUTF(name);
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'D':
        {
        Vector<String> v = new Vector<String>();
        int n = c.dis.readInt();
        for ( int i = 0; i < n; i++ )
          v.add(c.dis.readUTF());
        releaseConnection(c);
        return v.toArray(new String[0]);
        }
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...sfileModTs:2:
public long fileModTs(String name)
  throws FileNotFoundException, IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('T');
    c.dos.writeUTF(name);
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'T':
        {
        long ts = c.dis.readLong();
        releaseConnection(c);
        return ts;
        }
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...sreadFile:2:
public byte[] readFile(String name)
  throws FileNotFoundException, IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('R');
    c.dos.writeUTF(name);
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'R':
        {
        int n = c.dis.readInt();
        byte[] b = new byte[n];
        c.dis.readFully(b);
        releaseConnection(c);
        return b;
        }
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...swriteFile:2:
public boolean writeFile(String name, byte[] b)
  throws FileNotFoundException, IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('W');
    c.dos.writeUTF(name);
    c.dos.writeInt(b.length);
    c.dos.write(b);
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'W':
        releaseConnection(c);
        return true;
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...sdeleteFile:2:
public boolean deleteFile(String name)
  throws FileNotFoundException, IOException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('Z');
    c.dos.writeUTF(name);
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'Z':
        releaseConnection(c);
        return true;
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...sfindXPath:2:
// Calling
//   findXPath("/authors", ".*\\.xml", true, true, "//book[@genre='sci-fi']", null)
// could return
//   <?xml version="1.0"?>
//   <r:files xmlns:r="http://www.nyangau.org/rep">
//     <r:file name="/authors/asimov.xml">
//       <book genre="sci-fi">I Robot</book>
//       <book genre="sci-fi">Foundation</book>
//     </r:file>
//     <r:file name="/authors/clarke.xml">
//       <book genre="sci-fi">Rendevous with Rama</book>
//     </r:file>
//   </r:files>
public String findXPath(
  String name,      // can be file or directory
  String filter,    // can be regexp to match against file basename, or null
  boolean recurse,  // do subdirectories too
  boolean content,  // just list matching files, or include content too
  String xpath,     // XPath expression to use
  String nsprefixes // Namespace prefixes, or null
  )
  throws FileNotFoundException, IOException, PatternSyntaxException, XPathExpressionException
  {
  Connection c = acquireConnection();
  try
    {
    c.dos.writeByte('X');
    c.dos.writeUTF(name);
    c.dos.writeUTF(filter!=null?filter:"");
    c.dos.writeByte(recurse?1:0);
    c.dos.writeByte(content?1:0);
    c.dos.writeUTF(xpath);
    c.dos.writeUTF(nsprefixes!=null?nsprefixes:"");
    c.dos.flush();
    switch ( c.dis.readUnsignedByte() )
      {
      case 'N':
        releaseConnection(c);
        c = null;
        throw new FileNotFoundException(name+" in repository");
      case 'I':
        releaseConnection(c);
        c = null;
        throw new IOException(name+" in repository");
      case 'P':
        releaseConnection(c);
        c = null;
        throw new PatternSyntaxException("bad", filter, -1);
      case 'Q':
        releaseConnection(c);
        c = null;
        throw new XPathExpressionException(xpath);
      case 'X':
        {
        int l = c.dis.readInt();
        byte[] b = new byte[l];
        c.dis.readFully(b);
        releaseConnection(c);
        return new String(b, "UTF-8");
        }
      default:
        releaseConnection(c);
        c = null;
        throw new IOException("client/server mismatch?");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      c.disconnect();
    throw e;
    }
  }
//...e
//...sdebugCommand:2:
public boolean debugCommand(String[] args, int i)
  throws FileNotFoundException, IOException
  {
  String cmd = args[i];
  if ( cmd.equals("isFile") )
    System.out.println(isFile(args[i+1]));
  else if ( cmd.equals("isDirectory") )
    System.out.println(isDirectory(args[i+1]));
  else if ( cmd.equals("filesInDirectory") )
    {
    String[] a = filesInDirectory(args[i+1]);
    for ( int j = 0; j < a.length; j++ )
      System.out.println(a[j]);
    }
  else if ( cmd.equals("directoriesInDirectory") )
    {
    String[] a = directoriesInDirectory(args[i+1]);
    for ( int j = 0; j < a.length; j++ )
      System.out.println(a[j]);
    }
  else if ( cmd.equals("fileModTs") )
    {
    long ts = fileModTs(args[i+1]);
    System.out.println(new Date(ts));
    }
  else if ( cmd.equals("readFile") )
    {
    byte[] b = readFile(args[i+1]);
    System.out.write(b);
    }
  else if ( cmd.equals("writeFile") )
    System.out.println(writeFile(args[i+1], args[i+2].getBytes("UTF-8")));
  else if ( cmd.equals("deleteFile") )
    System.out.println(deleteFile(args[i+1]));
  else if ( cmd.equals("help") )
    {
    System.out.println("isFile name");
    System.out.println("isDirectory name");
    System.out.println("filesInDirectory name");
    System.out.println("directoriesInDirectory name");
    System.out.println("fileModTs name");
    System.out.println("readFile name");
    System.out.println("writeFile name text");
    System.out.println("deleteFile name");
    System.out.println("help");
    }
  else
    return false;
  return true;
  }
//...e
//...smain:2:
//...susage:2:
protected static void usage()
  {
  System.out.println("usage: SharedRepClient [flags] cmd {arg}");
  System.out.println("flags: -c connection  connection string (default: srep://localhost:6600)");
  System.out.println("       -P pwServer    password, or @file to read from file (default: password)");
  System.out.println("       cmd {arg}      command, with arguments");
  System.out.println("                      ls name");
  System.out.println("                      lsFile name");
  System.out.println("                      lsDir name");
  System.out.println("                      get name file");
  System.out.println("                      getFile name file");
  System.out.println("                      getDir name file");
  System.out.println("                      put file name"); 
  System.out.println("                      putFile file name"); 
  System.out.println("                      putDir file name"); 
  System.out.println("                      del name"); 
  System.out.println("                      delFile name"); 
  System.out.println("                      delDir name"); 
  System.out.println("                      xpath name pattern recurse content xpath nsctx"); 
  System.exit(1);
  }
//...e
//...slsFile\44\ lsDir\44\ ls:2:
public static void lsFileKnown(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  System.out.println(name);
  }

public static void lsDirKnown(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  String sep = name.endsWith("/") ? "" : "/";
  String[] fns = c.filesInDirectory(name);
  for ( int i = 0; i < fns.length; i++ )
    lsFileKnown(c, name+sep+fns[i]);
  String[] dns = c.directoriesInDirectory(name);
  for ( int i = 0; i < dns.length; i++ )
    lsDirKnown(c, name+sep+dns[i]);
  }

public static void lsFile(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  if ( c.isFile(name) )
    lsFileKnown(c, name);
  }

public static void lsDir(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  if ( c.isDirectory(name) )
    lsDirKnown(c, name);
  }

public static void ls(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  lsFile(c, name);
  lsDir(c, name);
  }
//...e
//...sgetFile\44\ getDir\44\ get:2:
public static void getFileKnown(SharedRepClient c, String name, String fn)
  throws FileNotFoundException, IOException
  {
  System.out.println(name+" -E-> "+fn);
  File pf = new File(fn).getParentFile();
  if ( pf != null )
    pf.mkdirs();
  byte[] b = c.readFile(name);
  FileOutputStream fos = new FileOutputStream(fn);
  fos.write(b);
  fos.close();
  }

public static void getDirKnown(SharedRepClient c, String name, String fn)
  throws FileNotFoundException, IOException
  {
  if ( c.isDirectory(name) )
    {
    String sep = name.endsWith("/") ? "" : "/";
    String fsep = fn.endsWith(File.separator) ? "" : File.separator;
    String[] fns = c.filesInDirectory(name);
    for ( int i = 0; i < fns.length; i++ )
      getFileKnown(c, name+sep+fns[i], fn+fsep+fns[i]);
    String[] dns = c.directoriesInDirectory(name);
    for ( int i = 0; i < dns.length; i++ )
      getDirKnown(c, name+sep+dns[i], fn+fsep+dns[i]);
    }
  }

public static void getFile(SharedRepClient c, String name, String fn)
  throws FileNotFoundException, IOException
  {
  if ( c.isFile(name) )
    getFileKnown(c, name, fn);
  }

public static void getDir(SharedRepClient c, String name, String fn)
  throws FileNotFoundException, IOException
  {
  if ( c.isDirectory(name) )
    getDirKnown(c, name, fn);
  }

public static void get(SharedRepClient c, String name, String fn)
  throws FileNotFoundException, IOException
  {
  getFile(c, name, fn);
  getDir(c, name, fn);
  }
//...e
//...sputFile\44\ putDir\44\ put:2:
public static void putFileKnown(SharedRepClient c, String fn, String name)
  throws FileNotFoundException, IOException
  {
  File f = new File(fn);
  System.out.println(fn+" -I-> "+name);
  FileInputStream fis = new FileInputStream(fn);
  long len = f.length();
  byte[] b = new byte[(int)len];
  fis.read(b);
  fis.close();
  c.writeFile(name, b);
  }

public static void putDirKnown(SharedRepClient c, String fn, String name)
  throws FileNotFoundException, IOException
  {
  File f = new File(fn);
  String sep = name.endsWith("/") ? "" : "/";
  String fsep = fn.endsWith(File.separator) ? "" : File.separator;
  String[] list = f.list();
  for ( int i = 0; i < list.length; i++ )
    {
    File nf = new File(f, list[i]);
    if ( nf.isFile() )
      putFileKnown(c, fn+fsep+list[i], name+sep+list[i]);
    }
  for ( int i = 0; i < list.length; i++ )
    {
    File nf = new File(f, list[i]);
    if ( nf.isDirectory() )
      putDirKnown(c, fn+fsep+list[i], name+sep+list[i]);
    }
  }

public static void putFile(SharedRepClient c, String fn, String name)
  throws FileNotFoundException, IOException
  {
  File f = new File(fn);
  if ( f.isFile() )
    putFileKnown(c, fn, name);
  }

public static void putDir(SharedRepClient c, String fn, String name)
  throws FileNotFoundException, IOException
  {
  File f = new File(fn);
  if ( f.isDirectory() )
    putDirKnown(c, fn, name);
  }

public static void put(SharedRepClient c, String fn, String name)
  throws FileNotFoundException, IOException
  {
  putFile(c, fn, name);
  putDir(c, fn, name);
  }
//...e
//...sdelFile\44\ delDir\44\ del:2:
public static void delFileKnown(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  System.out.println(name);
  c.deleteFile(name);
  }

public static void delDirKnown(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  String sep = name.endsWith("/") ? "" : "/";
  String[] fns = c.filesInDirectory(name);
  for ( int i = 0; i < fns.length; i++ )
    delFileKnown(c, name+sep+fns[i]);
  String[] dns = c.directoriesInDirectory(name);
  for ( int i = 0; i < dns.length; i++ )
    delDirKnown(c, name+sep+dns[i]);
  }

public static void delFile(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  if ( c.isFile(name) )
    delFileKnown(c, name);
  }

public static void delDir(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  if ( c.isDirectory(name) )
    delDirKnown(c, name);
  }

public static void del(SharedRepClient c, String name)
  throws FileNotFoundException, IOException
  {
  delFile(c, name);
  delDir(c, name);
  }
//...e

public static void main(String[] args)
  {
  try
    {
    String connection = "srep://localhost:6600";
    String pwServer = "password";
    boolean debug = false;

//...sparse arguments:6:
int i;
for ( i = 0; i < args.length; i++ )
  if ( args[i].charAt(0) != '-' )
    break;
  else if ( args[i].charAt(1) == '-' )
    { ++i; break; }
  else
    switch ( args[i].charAt(1) )
      {
      case 'c':
        if ( ++i == args.length )
          usage();
        connection = args[i];
        break;
      case 'P':
        if ( ++i == args.length )
          usage();
        pwServer = args[i];
        break;
      case 'd':
        debug = true;
        break;
      default:
        usage();
        break;
      }

if ( pwServer.startsWith("@") )
  {
  FileReader fr = new FileReader(pwServer.substring(1));
  BufferedReader br = new BufferedReader(fr);
  pwServer = br.readLine();
  }
//...e

    SharedRepClient c = new SharedRepClient(connection, pwServer);

    if ( i+2 == args.length && args[i].equals("ls") )
      ls(c, args[i+1]);
    else if ( i+2 == args.length && args[i].equals("lsFile") )
      lsFile(c, args[i+1]);
    else if ( i+2 == args.length && args[i].equals("lsDir") )
      lsDir(c, args[i+1]);
    else if ( i+3 == args.length && args[i].equals("get") )
      get(c, args[i+1], args[i+2]);
    else if ( i+3 == args.length && args[i].equals("getFile") )
      getFile(c, args[i+1], args[i+2]);
    else if ( i+3 == args.length && args[i].equals("getDir") )
      getDir(c, args[i+1], args[i+2]);
    else if ( i+3 == args.length && args[i].equals("put") )
      put(c, args[i+1], args[i+2]);
    else if ( i+3 == args.length && args[i].equals("putFile") )
      putFile(c, args[i+1], args[i+2]);
    else if ( i+3 == args.length && args[i].equals("putDir") )
      putDir(c, args[i+1], args[i+2]);
    else if ( i+2 == args.length && args[i].equals("del") )
      del(c, args[i+1]);
    else if ( i+2 == args.length && args[i].equals("delFile") )
      delFile(c, args[i+1]);
    else if ( i+2 == args.length && args[i].equals("delDir") )
      delDir(c, args[i+1]);
    else if ( i+7 == args.length && args[i].equals("xpath") )
      System.out.print(
        c.findXPath(
          args[i+1],
          args[i+2].equals("-") ? null : args[i+2],
          args[i+3].equalsIgnoreCase("true"),
          args[i+4].equalsIgnoreCase("true"),
          args[i+5],
          args[i+6].equals("-") ? null : args[i+6]
          )
        );
    else if ( debug )
      {
      if ( ! c.debugCommand(args, i) )
        System.out.println("use help command to see debug commands");
      }
    else
      usage();
    }
  catch ( Exception e )
    {
    e.printStackTrace();
    }
  }
//...e
  }
