//
// SharedRepLocation.java - Manage the interface to other cluster locations
//

package nyangau.rep;

//...simports:0:
import java.util.*;
import java.util.logging.*;
import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import javax.net.*;
import javax.net.ssl.*;
import javax.security.auth.x500.*;
import javax.xml.xpath.XPathExpressionException;
import java.lang.management.*;
import javax.management.*;
//...e

public class SharedRepLocation implements SharedRepLocationMBean
  {
  private static Logger logger = Logger.getLogger(SharedRepLocation.class.getName());
  public int connNextOutbound;
  public int connNextInbound;
  protected SharedRep srep;
  protected int loc;
  protected String host;
  protected int port;
  protected Object oSyncDirty;
  protected Set<String> dirty;
  protected Object oSyncMap;
  protected Map<String,MapValue> map;
  protected Sender sender;
  protected Object oSyncReceiver;
  protected Receiver receiver;
  protected static final long T_MAX_SKEW = 300000; // 5mins
  protected static final long T_SKEW_RETRY_DELAY = 60000; // 1min
  protected Object oSyncPrev;
  protected Connection cPrev;
  protected Reaper reaper;
  protected boolean connectedOutbound;
  protected int connectedInbound;
//...sclass Connection:2:
public class Connection
  {
  protected Socket s;
  public DataOutputStream dos;
  public DataInputStream  dis;
  public long lastUsed;
  public Connection()
    throws IOException
    {
    s   = srep.connectTo(host, port);
    dis = new DataInputStream (s.getInputStream ());
    dos = new DataOutputStream(s.getOutputStream());
    }
  public void disconnect()
    throws IOException
    {
    dis.close();
    dos.close();
    s.close();
    }
  }
//...e
//...sclass Reaper:2:
public class Reaper extends Thread
  {
  public static final long T_LINGER = 10000;
  public static final long T_CHECK  =  5000;
  public void run()
    {
    for ( ;; )
      {
      try
        {
        sleep(T_CHECK);
        }
      catch ( InterruptedException e )
        {
        }
      synchronized ( oSyncPrev )
        {
        long t = System.currentTimeMillis();
        if ( cPrev != null &&
             cPrev.lastUsed + T_LINGER < t )
          {
          try
            {
            cPrev.disconnect();
            }
          catch ( IOException e )
            {
            }
          cPrev = null;
          reaper = null;
          return; // Reaper ends
          }
        }
      }
    }
  }
//...e
//...sclass Sender:2:
public class Sender extends Thread
  {
  protected static final long T_CONNECT   =  5000;
  protected static final long T_AUTH      = 60000;
  protected static final long T_CLEAN     =  5000;
  protected static final long T_HEARTBEAT = 60000;

  public void run()
    {
    for ( ;; )
      {
      Connection c = null;
      try
        {
        c = new Connection();
        connNextOutbound++;
        logger.info("Sender: connected to "+loc);
        connectedOutbound = true;
        c.dos.writeUTF(srep.getPassword());
        c.dos.flush();
        int rc = c.dis.readUnsignedByte();
        if ( rc == 'S' )
          {
          logger.warning("Sender: authentication error talking to "+loc);
          connectedOutbound = false;
          try
            {
            sleep(T_AUTH);
            }
          catch ( InterruptedException e )
            {
            }
          throw new IOException();
          }
        c.dos.writeByte('M');          // we are sending metadata
        c.dos.writeInt(srep.getLoc()); // from our location
        c.dos.writeLong(System.currentTimeMillis()+srep.getTimeAdjustment());
        c.dos.flush();
        synchronized ( oSyncDirty )
          {
          dirty.addAll( srep.getMapKeys() ); // need to send everything
          }
        long idle = 0;
        for ( ;; )
          {
          String name = null;
          synchronized ( oSyncDirty )
            {
            Iterator<String> it = dirty.iterator();
            if ( it.hasNext() )
              {
              name = it.next();
              }
            }
          if ( name != null )
            {
            MapValue mvalue = srep.getMapValue(name);
            c.dos.writeUTF(name);
            c.dos.writeByte(mvalue.ft);
            c.dos.writeInt(mvalue.loc);
            c.dos.writeLong(mvalue.ts);
            c.dos.flush();
            c.dis.readUnsignedByte(); // any confirmation is good
            MapValue mvalue2 = srep.getMapValue(name);
            if ( mvalue.ft  == mvalue2.ft  &&
                 mvalue.loc == mvalue2.loc &&
                 mvalue.ts  == mvalue2.ts  )
              synchronized ( oSyncDirty )
                {
                dirty.remove(name);
                }
            idle = 0;
            }
          else
            {
            try
              {
              sleep(T_CLEAN);
              }
            catch ( InterruptedException e )
              {
              }
            idle += T_CLEAN;
            if ( idle >= T_HEARTBEAT )
              {
              c.dos.writeUTF(""); // heartbeat to keep socket alive
              c.dos.flush();
              c.dis.readUnsignedByte(); // any confirmation is good
              idle = 0;
              }
            }
          }
        }
      catch ( IOException e )
        {
        if ( c != null )
          {
          try
            {
            c.disconnect();
            }
          catch ( IOException e2 )
            {
            }
          logger.info("Sender: disconnected from "+loc);
          connectedOutbound = false;
          }
        try
          {
          sleep(T_CONNECT);
          }
        catch ( InterruptedException e2 )
          {
          }
        }
      }
    }
  }
//...e
//...sclass Receiver:2:
public class Receiver extends Thread
  {
  public int conn;
  protected Socket           s;
  protected DataInputStream  dis;
  protected DataOutputStream dos;
  public Receiver(Socket s, DataInputStream dis, DataOutputStream dos)
    {
    this.s   = s;
    this.dis = dis;
    this.dos = dos;
    conn = connNextInbound++;
    }
  protected void close()
    {
    if ( s != null )
      {
      try
        {
        s.close();
        s = null;
        }
      catch ( IOException e )
        {
        }
      }
    }
  public void run()
    {
    // password checked out ok
    // stream of data known to be a metadata stream
    // stream of data known to have originated from parent location
    try
      {
      long tsThere = dis.readLong();
      long tsHere = System.currentTimeMillis()+srep.getTimeAdjustment();
      long tsSkew = Math.abs(tsHere-tsThere);
      if ( tsSkew > T_MAX_SKEW )
        {
        logger.warning("Receiver: clocks too far apart "+loc+" #"+conn+" (skew="+tsSkew+"ms), disconnecting");
        try
          {
          sleep(T_SKEW_RETRY_DELAY); // don't reconnect straight away
          }
        catch ( InterruptedException e )
          {
          }
        receiver = null;
        close();
        return;
        }
      logger.info("Receiver: connected to "+loc+" #"+conn+" (skew="+tsSkew+"ms)");
      ++connectedInbound;
      for ( ;; )
        {
        String name = dis.readUTF();
        if ( name.equals("") )
          // a dummy notification to keep the connection alive
          // and to enable us to detect when it has failed
          ;
        else
          // a real notification
          {
          byte ftN = (byte) dis.readUnsignedByte();
          int locN = dis.readInt();
          long tsN = dis.readLong();
          MapValue mvalue = new MapValue(ftN, locN, tsN);
          synchronized ( oSyncMap )
            {
            map.put(name, mvalue);
            }
          srep.mapValueChanged(loc, conn, name);
          }
        dos.writeByte(0); // any byte is a confirmation
        dos.flush();
        }
      }
    catch ( IOException e )
      {
      synchronized ( oSyncReceiver )
        {
        receiver = null;
        close();
        logger.info("Receiver: disconnected from "+loc+" #"+conn);
        --connectedInbound;
        }
      }
    }
  public int getConn()
    {
    return conn;
    }
  public void kill()
    {
    close();
    logger.info("Receiver: closed prior connection from "+loc+" #"+conn);
    }
  }
//...e
//...sSharedRepLocation:2:
public SharedRepLocation(SharedRep srep, int loc, String host, int port)
  {
  connNextOutbound = 0;
  connNextInbound = 0;

  this.srep = srep;
  this.loc  = loc;
  this.host = host;
  this.port = port;

  connectedOutbound = false;
  connectedInbound  = 0;

  oSyncDirty = new Object();
  dirty = new HashSet<String>();

  oSyncMap = new Object();
  map = new HashMap<String,MapValue>();

  oSyncPrev = new Object();
  cPrev = null;
  reaper = null;

  oSyncReceiver = new Object();
  receiver = null;

  sender = new Sender();
  sender.start();

  // Register MBean so people can see how dirty we are
  try
    {
    MBeanServer mbs = ManagementFactory.getPlatformMBeanServer(); 
    mbs.registerMBean(this, new ObjectName("nyangau.rep:type=SharedRepLocation,loc="+loc));
    }
  catch ( MalformedObjectNameException e )
    {
    }
  catch ( InstanceAlreadyExistsException e )
    {
    }
  catch ( MBeanRegistrationException e )
    {
    }
  catch ( NotCompliantMBeanException e )
    {
    }
  }
//...e
//...sacquireConnection:2:
// Throws IOException if cannot connect
// Throws IOException if authentication error
// Returns connection otherwise

public Connection acquireConnection()
  throws IOException
  {
  synchronized ( oSyncPrev )
    {
    if ( cPrev != null )
      {
      Connection c = cPrev;
      cPrev = null;
      return c;
      }
    }
  Connection c = null;
  try
    {
    c = new Connection();
    c.dos.writeUTF(srep.getPassword());
    c.dos.flush();
    if ( c.dis.readUnsignedByte() == 'S' )
      {
      c.disconnect();
      throw new IOException("authentication failed");
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      try
        {
        c.disconnect();
        }
      catch ( IOException e2 )
        {
        }
    throw e;
    }
  return c;
  }
//...e
//...sreleaseConnection:2:
public void releaseConnection(Connection c)
  {
  c.lastUsed = System.currentTimeMillis();
  synchronized ( oSyncPrev )
    {
    cPrev = c;
    if ( reaper == null )
      {
      reaper = new Reaper();
      reaper.setDaemon(true);
      reaper.start();
      }
    }
  }
//...e
//...snewReceiver:2:
public void newReceiver(Socket s, DataInputStream dis, DataOutputStream dos)
  {
  synchronized ( oSyncReceiver )
    {
    if ( receiver != null )
      receiver.kill();
    receiver = new Receiver(s, dis, dos);
    receiver.start();
    }
  }
//...e
//...sgetReceiverConn:2:
public int getReceiverConn()
  {
  synchronized ( oSyncReceiver )
    {
    return ( receiver != null ) ? receiver.getConn() : -1;
    }
  }
//...e
//...smarkDirty:2:
public void markDirty(String d)
  {
  synchronized ( oSyncDirty )
    {
    dirty.add(d);
    }
  }
//...e
//...sgetMapValue:2:
public MapValue getMapValue(String name)
  {
  synchronized ( oSyncMap )
    {
    return map.get(name);
    }
  }
//...e
//...stoString:2:
public String toString()
  {
  String m;
  synchronized ( oSyncMap )
    {
    m = map.toString();
    }
  String d;
  synchronized ( oSyncDirty )
    {
    d = dirty.toString();
    }
  String r = ( receiver == null ) ? "==" : "!=";
  return "{loc="+loc+",host="+host+",port="+port+",map="+m+",dirty="+d+",receiver"+r+"null}";
  }
//...e
//...sgetLoc:2:
// part of SharedRepLocationMBean interface
public int getLoc()
  {
  return loc;
  }
//...e
//...sgetDirty:2:
// part of SharedRepLocationMBean interface
public int getDirty()
  {
  synchronized ( oSyncDirty )
    {
    return dirty.size();
    }
  }
//...e
//...sgetConnectedOutbound:2:
// part of SharedRepLocationMBean interface
public boolean getConnectedOutbound()
  { return connectedOutbound; }
//...e
//...sgetConnectedInbound:2:
// part of SharedRepLocationMBean interface
public boolean getConnectedInbound()
  { return connectedInbound > 0; }
//...e
//...sgetConnectionNumberOutbound:2:
public int getConnectionNumberOutbound()
  {
  return connNextOutbound;
  }
//...e
//...sgetConnectionNumberInbound:2:
public int getConnectionNumberInbound()
  {
  return connNextInbound;
  }
//...e
  }
